/* libhttpd.c - HTTP protocol library
**
** Copyright (C) 1995 by Jef Poskanzer <jef@acme.com>.  All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
** 
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/


#include "config.h"
#include "version.h"

#include <sys/types.h>
#include <sys/param.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <errno.h>
#include <fcntl.h>
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#ifdef HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# ifdef HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# ifdef HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# ifdef HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

#include "libhttpd.h"
#include "timers.h"
#include "match.h"
#include "tdate_parse.h"

#ifndef STDIN_FILENO
#define STDIN_FILENO 0
#endif
#ifndef STDOUT_FILENO
#define STDOUT_FILENO 1
#endif
#ifndef STDERR_FILENO
#define STDERR_FILENO 2
#endif


static void
child_handler( int sig )
    {
    pid_t pid;
    int status;

    /* Set up the signal handler again.  Don't need to do this on BSD
    ** systems, but it doesn't hurt.
    */
    (void) signal( SIGCHLD, child_handler );

    /* Reap defunct children until there aren't any more. */
    for (;;)
	{
#ifdef HAVE_WAITPID
	pid = waitpid( (pid_t) -1, &status, WNOHANG );
#else /* HAVE_WAITPID */
	pid = wait3( &status, WNOHANG, (struct rusage*) 0 );
#endif /* HAVE_WAITPID */
	if ( (int) pid == 0 )		/* none left */
	    break;
	if ( (int) pid < 0 )
	    {
	    if ( errno == EINTR )	/* because of ptrace */
		continue;
	    /* ECHILD shouldn't happen with the WNOHANG option, but with
	    ** some kernels it does anyway.  Ignore it.
	    */
	    if ( errno != ECHILD )
		syslog( LOG_ERR, "waitpid - %m" );
	    break;
	    }
	}
    }


static void
check_options( void )
    {
#if defined(TILDE_MAP_1) && defined(TILDE_MAP_2)
    syslog( LOG_CRIT, "both TILDE_MAP_1 and TILDE_MAP_2 are defined" );
    exit( 1 );
#endif /* both */

#if defined(LOG_REFERER_ALWAYS) && defined(LOG_REFERER_ON_ERRS)
    syslog( LOG_CRIT,
	"both LOG_REFERER_ALWAYS and LOG_REFERER_ON_ERRS are defined" );
    exit( 1 );
#endif /* both */

#if defined(LOG_USERAGENT_ALWAYS) && defined(LOG_USERAGENT_ON_ERRS)
    syslog( LOG_CRIT,
	"both LOG_USERAGENT_ALWAYS and LOG_USERAGENT_ON_ERRS are defined" );
    exit( 1 );
#endif /* both */
    }


static void
free_httpd_server( httpd_server* hs )
    {
    if ( hs->cwd != (char*) 0 )
	free( (void*) hs->cwd );
    if ( hs->cgi_pattern != (char*) 0 )
	free( (void*) hs->cgi_pattern );
    free( (void*) hs );
    }


httpd_server*
httpd_initialize( char* hostname, int port, char* cgi_pattern, char* cwd )
    {
    httpd_server* hs;
    int on;
    struct hostent* he;
    struct sockaddr_in sa;
    char* cp;

    check_options();

    /* Set up child-process reaper. */
    (void) signal( SIGCHLD, child_handler );

    hs = NEW( httpd_server, 1 );
    if ( hs == (httpd_server*) 0 )
	{
	syslog( LOG_CRIT, "out of memory" );
	return (httpd_server*) 0;
	}
    if ( hostname == (char*) 0 )
	hs->hostname = (char*) 0;
    else
	hs->hostname = strdup( hostname );
    hs->port = port;
    if ( cgi_pattern == (char*) 0 )
	hs->cgi_pattern = (char*) 0;
    else
	{
	/* Nuke any leading slashes. */
	if ( cgi_pattern[0] == '/' )
	    ++cgi_pattern;
	hs->cgi_pattern = strdup( cgi_pattern );
	if ( hs->cgi_pattern == (char*) 0 )
	    {
	    syslog( LOG_CRIT, "out of memory" );
	    return (httpd_server*) 0;
	    }
	/* Nuke any leading slashes in the cgi pattern. */
	while ( ( cp = strstr( hs->cgi_pattern, "|/" ) ) != (char*) 0 )
	    (void) strcpy( cp + 1, cp + 2 );
	}
    hs->cwd = strdup( cwd );
    if ( hs->cwd == (char*) 0 )
	{
	syslog( LOG_CRIT, "out of memory" );
	return (httpd_server*) 0;
	}

    /* Create socket. */
    hs->fd = socket( AF_INET, SOCK_STREAM, 0 );
    if ( hs->fd < 0 )
	{
	syslog( LOG_CRIT, "socket - %m" );
	free_httpd_server( hs );
	return (httpd_server*) 0;
	}

    /* Allow reuse of local addresses. */
    on = 1;
    if ( setsockopt(
	     hs->fd, SOL_SOCKET, SO_REUSEADDR, (char*) &on, sizeof(on) ) < 0 )
        syslog( LOG_CRIT, "setsockopt SO_REUSEADDR - %m" );

    /* Bind to it. */
    memset( (char*) &sa, 0, sizeof(sa) );
    sa.sin_family = AF_INET;
    if ( hs->hostname == (char*) 0 )
	sa.sin_addr.s_addr = htonl( INADDR_ANY );
    else
	{
	sa.sin_addr.s_addr = inet_addr( hs->hostname );
	if ( (int) sa.sin_addr.s_addr == -1 )
	    {
	    he = gethostbyname( hs->hostname );
	    if ( he == (struct hostent*) 0 )
		{
		syslog( LOG_CRIT, "gethostbyname - %m" );
		free_httpd_server( hs );
		return (httpd_server*) 0;
		}
	    if ( he->h_addrtype != AF_INET ||
		 he->h_length != sizeof(sa.sin_addr) )
		{
		syslog(
		    LOG_CRIT, "%.80s - non-IP network address", hs->hostname );
		free_httpd_server( hs );
		return (httpd_server*) 0;
		}
	    (void) memcpy(
		&sa.sin_addr.s_addr, he->h_addr, sizeof(sa.sin_addr) );
	    }
	}
    sa.sin_port = htons( hs->port );
    hs->host_addr = sa.sin_addr;
    if ( bind( hs->fd, (struct sockaddr*) &sa, sizeof(sa) ) < 0 )
	{
	syslog( LOG_CRIT, "bind - %m" );
	(void) close( hs->fd );
	free_httpd_server( hs );
	return (httpd_server*) 0;
	}

    /* Start a listen going. */
    if ( listen( hs->fd, LISTEN_BACKLOG ) < 0 )
	{
	syslog( LOG_CRIT, "listen - %m" );
	(void) close( hs->fd );
	free_httpd_server( hs );
	return (httpd_server*) 0;
	}

    /* Done initializing. */
    if ( hs->hostname == (char*) 0 )
	syslog( LOG_INFO, "%s starting on port %d", SERVER_SOFTWARE, hs->port );
    else
	syslog(
	    LOG_INFO, "%s starting on %.80s, port %d", SERVER_SOFTWARE,
	    inet_ntoa( hs->host_addr ), hs->port );
    return hs;
    }


void
httpd_terminate( httpd_server* hs )
    {
    (void) close( hs->fd );
    free_httpd_server( hs );
    }


static char* ok200title = "OK";

static char* err302title = "Found";
static char* err302form = "The actual URL is '%.80s'.\n";

static char* err304title = "Not Modified";

static char* err400title = "Bad Request";
static char* err400form =
    "Your request '%.80s' has bad syntax or is inherently impossible to satisfy.\n";

static char* err403title = "Forbidden";
static char* err403form =
    "You do not have permission to get URL '%.80s' from this server.\n";

static char* err404title = "Not Found";
static char* err404form =
    "The requested URL '%.80s' was not found on this server.\n";

static char* err408title = "Request Timeout";
static char* err408form =
    "No request appeared within a reasonable time period.\n";

static char* err500title = "Internal Error";
static char* err500form =
    "There was an unusual problem serving the requested URL '%.80s'.\n";

static char* err501title = "Not Implemented";
static char* err501form =
    "The requested method '%.80s' is not implemented by this server.\n";

char* httpd_err503title = "Service Temporarily Overloaded";
char* httpd_err503form =
    "The requested URL '%.80s' is temporarily overloaded.  Please try again later.\n";


static void
send_mime( httpd_conn* hc, int status, char* title, char* encodings, char* extraheads, char* type, int length, time_t mod )
    {
    char* enc_head = "\nContent-encoding: ";
    char* mime_len_head =
     "HTTP/1.0 %d %s\nServer: %s%s%s%s%s\nContent-type: %s\nContent-length: %d\nLast-modified: %s\n\n";
    char* mime_nolen_head =
     "HTTP/1.0 %d %s\nServer: %s%s%s%s%s\nContent-type: %s\nLast-modified: %s\n\n";
    char tbuf[100];
    char buf[1000];

    hc->status = status;
    hc->bytes = length;
    if ( hc->mime_flag )
	{
	if ( mod == (time_t) 0 )
	    mod = time( (time_t*) 0 );
	(void) strftime(
	    tbuf, sizeof(tbuf), "%a, %d-%b-%y %H:%M:%S GMT", gmtime( &mod ) );
	if ( length >= 0 )
	    (void) sprintf(
		buf, mime_len_head, status, title, SERVER_SOFTWARE,
		encodings[0] != '\0' ? enc_head : "",
		encodings[0] != '\0' ? encodings : "",
		extraheads[0] != '\0' ? "\n" : "", extraheads,
		type, length, tbuf );
	else
	    (void) sprintf(
		buf, mime_nolen_head, status, title, SERVER_SOFTWARE,
		encodings[0] != '\0' ? enc_head : "",
		encodings[0] != '\0' ? encodings : "",
		extraheads[0] != '\0' ? "\n" : "", extraheads,
		type, tbuf );
	(void) write( hc->fd, buf, strlen( buf ) );
	}
    }


static int
send_address( httpd_conn* hc )
    {
    char buf[1000];
    int len;

    (void) sprintf( buf,
	"<ADDRESS><A HREF=\"%s\">%s</A></ADDRESS>\n",
	SERVER_ADDRESS, SERVER_SOFTWARE );
    len = strlen( buf );
    (void) write( hc->fd, buf, len );
    return len;
    }


static void
realloc_str( char** strP, int* maxsizeP, int size )
    {
    if ( *maxsizeP == 0 )
	{
	*maxsizeP = MAX( 200, size );	/* arbitrary */
	*strP = NEW( char, *maxsizeP + 1 );
	}
    else if ( size > *maxsizeP )
	{
	*maxsizeP = MAX( *maxsizeP * 2, size * 5 / 4 );
	*strP = RENEW( *strP, char, *maxsizeP + 1 );
	}
    else
	return;
    if ( *strP == (char*) 0 )
	{
	syslog( LOG_ERR, "out of memory" );
	exit( 1 );
	}
    }


static void
send_response( httpd_conn* hc, int status, char* title, char* extrahead, char* form, char* arg )
    {
    char buf[1000];

    send_mime( hc, status, title, "", extrahead, "text/html", -1, 0 );
    (void) sprintf(
	buf, "<HTML><HEAD><TITLE>%d %s</TITLE></HEAD>\n<BODY><H2>%d %s</H2>\n",
	status, title, status, title );
    (void) write( hc->fd, buf, strlen( buf ) );
    (void) sprintf( buf, form, arg );
    (void) write( hc->fd, buf, strlen( buf ) );
    (void) sprintf( buf, "<HR>\n" );
    (void) write( hc->fd, buf, strlen( buf ) );
    (void) send_address( hc );
    (void) sprintf( buf, "</BODY></HTML>\n" );
    (void) write( hc->fd, buf, strlen( buf ) );
#ifdef LOG_REFERER_ON_ERRS
    if ( hc->referer[0] != '\0' )
	syslog(
	    LOG_INFO, "referer %.80s -> %.80s", hc->referer, hc->encodedurl );
#endif /* LOG_REFERER_ON_ERRS */
#ifdef LOG_USERAGENT_ON_ERRS
    if ( hc->useragent[0] != '\0' )
	syslog( LOG_INFO, "user-agent %.80s", hc->useragent );
#endif /* LOG_USERAGENT_ON_ERRS */
    }


void
httpd_send_err( httpd_conn* hc, int status, char* title, char* form, char* arg )
    {
    send_response( hc, status, title, "", form, arg );
    }


static void
send_dirredirect( httpd_conn* hc )
    {
    static char* location;
    static char* header;
    static int maxlocation = 0, maxheader = 0;
    static char headstr[] = "Location: ";

    realloc_str( &location, &maxlocation, strlen( hc->encodedurl ) + 1 );
    (void) sprintf( location, "%s/", hc->encodedurl );
    realloc_str( &header, &maxheader, sizeof(headstr) + strlen( location ) );
    (void) sprintf( header, "%s%s", headstr, location );
    send_response( hc, 302, err302title, header, err302form, location );
    }


char*
httpd_method_str( int method )
    {
    switch ( method )
	{
	case METHOD_GET: return "GET";
	case METHOD_HEAD: return "HEAD";
	case METHOD_POST: return "POST";
	default: return (char*) 0;
	}
    }


static int
is_hexit( char c )
    {
    if ( strchr( "0123456789abcdefABCDEF", c ) != (char*) 0 )
	return 1;
    return 0;
    }


static int
hexit( char c )
    {
    if ( c >= '0' && c <= '9' )
	return c - '0';
    if ( c >= 'a' && c <= 'f' )
	return c - 'a' + 10;
    if ( c >= 'A' && c <= 'F' )
	return c - 'A' + 10;
    return 0;		/* shouldn't happen, we're guarded by is_hexit() */
    }


/* Copies and decodes a string.  It's ok for from and to to be the
** same string.
*/
static void
strdecode( char* to, char* from )
    {
    for ( ; *from != '\0'; ++to, ++from )
	{
	if ( from[0] == '%' && is_hexit( from[1] ) && is_hexit( from[2] ) )
	    {
	    *to = hexit( from[1] ) * 16 + hexit( from[2] );
	    from += 2;
	    }
	else
	    *to = *from;
	}
    *to = '\0';
    }


static void
alarm_handler( int sig )
    {
#ifndef SV_INTERRUPT
    /* Set up the signal handler again.  Don't do this on BSD
    ** systems, because SV_INTERRUPT would be cleared.
    */
    (void) signal( SIGALRM, alarm_handler );
#endif /* SV_INTERRUPT */
    }


int
httpd_get_nfiles( void )
    {
    static int inited = 0;
    static int n;

    if ( ! inited )
	{
#ifdef RLIMIT_NOFILE
	struct rlimit rl;
	if ( getrlimit( RLIMIT_NOFILE, &rl ) < 0 )
	    {
	    syslog( LOG_ERR, "getrlimit - %m" );
	    exit( 1 );
	    }
	if ( rl.rlim_max == RLIM_INFINITY )
	    rl.rlim_cur = 4096;		/* arbitrary */
	else
	    rl.rlim_cur = rl.rlim_max;
	if ( setrlimit( RLIMIT_NOFILE, &rl ) < 0 )
	    {
	    syslog( LOG_ERR, "setrlimit - %m" );
	    exit( 1 );
	    }
	n = rl.rlim_cur;
#else /* RLIMIT_NOFILE */
	n = getdtablesize();
#endif /* RLIMIT_NOFILE */
	inited = 1;
	}
    return n;
    }


/* Map a ~username/whatever URL into something else.  Two different ways. */
static int
tilde_map( httpd_conn* hc )
    {
#if defined(TILDE_MAP_1) || defined(TILDE_MAP_2)
    static char* temp;
    static int maxtemp = 0;
#endif
#ifdef TILDE_MAP_1
    /* Map ~username to <prefix>/username. */
    int len;
    static char* prefix = TILDE_MAP_1;

    len = strlen( hc->expnfilename ) - 1;
    realloc_str( &temp, &maxtemp, len );
    (void) strcpy( temp, &hc->expnfilename[1] );
    realloc_str(
	&hc->expnfilename, &hc->maxexpnfilename, strlen( prefix ) + 1 + len );
    (void) strcpy( hc->expnfilename, prefix );
    if ( prefix[0] != '\0' )
	(void) strcat( hc->expnfilename, "/" );
    (void) strcat( hc->expnfilename, temp );
#endif /* TILDE_MAP_1 */

#ifdef TILDE_MAP_2
    /* Map ~username to <user's homedir>/<postfix>. */
    static char* postfix = TILDE_MAP_2;
    char* cp;
    struct passwd* pw;

    /* Get the username. */
    realloc_str( &temp, &maxtemp, strlen( hc->expnfilename ) - 1 );
    (void) strcpy( temp, &hc->expnfilename[1] );
    cp = strchr( temp, '/' );
    if ( cp != (char*) 0 )
	*cp++ = '\0';
    else
	cp = "";
    
    /* Get the passwd entry. */
    pw = getpwnam( temp );
    if ( pw == (struct passwd*) 0 )
	return 0;
    
    /* Set up altdir. */
    realloc_str(
	&hc->altdir, &hc->maxaltdir,
	strlen( pw->pw_dir ) + 1 + strlen( postfix ) );
    (void) strcpy( hc->altdir, pw->pw_dir );
    if ( postfix[0] != '\0' )
	{
	(void) strcat( hc->altdir, "/" );
	(void) strcat( hc->altdir, postfix );
	}

    /* And the filename becomes altdir plus the post-~ part of the original. */
    realloc_str(
	&hc->expnfilename, &hc->maxexpnfilename,
	strlen( hc->altdir ) + 1 + strlen( cp ) );
    (void) sprintf( hc->expnfilename, "%s/%s", hc->altdir, cp );
#endif /* TILDE_MAP_2 */
    return 1;
    }


/* Expands all symlinks in the given filename, eliding ..'s.  Returns the
** expanded path (pointer to static string), or (char*) 0 on errors.
** Also returns, in the string pointed to by restP, any trailing parts of
** the path that don't exist.
**
** This is a fairly nice little routine.  It handles any size filenames
** without excessive mallocs.
*/
static char*
expand_symlinks( char* path, char** restP )
    {
    static char* checked;
    static char* rest;
    char link[5000];
    static int maxchecked = 0, maxrest = 0;
    int checkedlen, restlen, linklen, prevcheckedlen, prevrestlen, nlinks, i;
    char* r;
    char* cp1;
    char* cp2;

    /* Start out with nothing in checked and the whole filename in rest. */
    realloc_str( &checked, &maxchecked, 0 );
    checked[0] = '\0';
    checkedlen = 0;
    restlen = strlen( path );
    realloc_str( &rest, &maxrest, restlen );
    (void) strcpy( rest, path );
    if ( rest[restlen - 1] == '/' )
	rest[--restlen] = '\0';		/* trim trailing slash */
    r = rest;
    nlinks = 0;

    /* While there are still components to check... */
    while ( restlen > 0 )
	{
	/* Save current checkedlen in case we get a symlink.  Save current
	** restlen in case we get a non-existant component.
	*/
	prevcheckedlen = checkedlen;
	prevrestlen = restlen;

	/* Grab one component from r and transfer it to checked. */
	cp1 = strchr( r, '/' );
	if ( cp1 != (char*) 0 )
	    {
	    i = cp1 - r;
	    if ( i == 0 )
		{
		/* Special case for absolute paths. */
		realloc_str( &checked, &maxchecked, checkedlen + 1 );
		(void) strncpy( &checked[checkedlen], r, 1 );
		checkedlen += 1;
		}
	    else if ( strncmp( r, "..", MAX( i, 2 ) ) == 0 )
		{
		/* Ignore ..'s that go above the start of the path. */
		if ( checkedlen != 0 )
		    {
		    cp2 = strrchr( checked, '/' );
		    if ( cp2 == (char*) 0 )
			checkedlen = 0;
		    else if ( cp2 == checked )
			checkedlen = 1;
		    else
			checkedlen = cp2 - checked;
		    }
		}
	    else
		{
		realloc_str( &checked, &maxchecked, checkedlen + 1 + i );
		if ( checkedlen > 0 && checked[checkedlen-1] != '/' )
		    checked[checkedlen++] = '/';
		(void) strncpy( &checked[checkedlen], r, i );
		checkedlen += i;
		}
	    checked[checkedlen] = '\0';
	    r += i + 1;
	    restlen -= i + 1;
	    }
	else
	    {
	    /* No slashes remaining, r is all one component. */
	    if ( strcmp( r, ".." ) == 0 )
		{
		/* Ignore ..'s that go above the start of the path. */
		if ( checkedlen != 0 )
		    {
		    cp2 = strrchr( checked, '/' );
		    if ( cp2 == (char*) 0 )
			checkedlen = 0;
		    else if ( cp2 == checked )
			checkedlen = 1;
		    else
			checkedlen = cp2 - checked;
		    checked[checkedlen] = '\0';
		    }
		}
	    else
		{
		realloc_str( &checked, &maxchecked, checkedlen + 1 + restlen );
		if ( checkedlen > 0 && checked[checkedlen-1] != '/' )
		    checked[checkedlen++] = '/';
		(void) strcpy( &checked[checkedlen], r );
		checkedlen += restlen;
		}
	    r += restlen;
	    restlen = 0;
	    }

	/* Try reading the current filename as a symlink */
	linklen = readlink( checked, link, sizeof(link) );
	if ( linklen == -1 )
	    {
	    if ( errno == EINVAL )
		continue;		/* not a symlink */
	    if ( errno == EACCES || errno == ENOENT || errno == ENOTDIR )
		{
		/* That last component was bogus.  Restore and return. */
		*restP = r - ( prevrestlen - restlen );
		checked[prevcheckedlen] = '\0';
		return checked;
		}
	    syslog( LOG_ERR, "readlink - %m" );
	    return (char*) 0;
	    }
	++nlinks;
	if ( nlinks > MAX_LINKS )
	    {
	    syslog( LOG_ERR, "too many symlinks" );
	    return (char*) 0;
	    }
	link[linklen] = '\0';
	if ( link[linklen - 1] == '/' )
	    link[--linklen] = '\0';	/* trim trailing slash */

	/* Insert the link contents in front of the rest of the filename. */
	if ( restlen != 0 )
	    {
	    (void) strcpy( rest, r );
	    realloc_str( &rest, &maxrest, restlen + linklen + 1 );
	    for ( i = restlen; i >= 0; --i )
		rest[i + linklen + 1] = rest[i];
	    (void) strcpy( rest, link );
	    rest[linklen] = '/';
	    restlen += linklen + 1;
	    r = rest;
	    }
	else
	    {
	    /* There's nothing left in the filename, so the link contents
	    ** becomes the rest.
	    */
	    realloc_str( &rest, &maxrest, linklen );
	    (void) strcpy( rest, link );
	    restlen = linklen;
	    r = rest;
	    }

	/* Re-check this component. */
	checkedlen = prevcheckedlen;
	checked[checkedlen] = '\0';
	}

    /* Ok. */
    *restP = r;
    return checked;
    }


int
httpd_get_conn( httpd_server* hs, httpd_conn* hc )
    {
    struct sockaddr_in sin;
    static int inited_sig = 0;
    int sz;
    char buf[5000];
    char* method_str;
    char* url;
    char* protocol;
    char* eol;
    char* cp;
    char* pi;

    if ( ! inited_sig )
	{
#ifdef HAVE_SIGVEC
	/* In BSD you have to explicitly set a signal to interrupt syscalls. */
	struct sigvec sv;
	sv.sv_handler = alarm_handler;
	sv.sv_mask = 0;
	sv.sv_flags = SV_INTERRUPT;
	if ( sigvec( SIGALRM, &sv, (struct sigvec*) 0 ) < 0 )
	    {
	    syslog( LOG_ERR, "sigvec - %m" );
	    exit( 1 );
	    }
#else /* HAVE_SIGVEC */
	/* In SYSV signals always interrupt system calls. */
	(void) signal( SIGALRM, alarm_handler );
#endif /* HAVE_SIGVEC */
	inited_sig = 1;
	}

    if ( ! hc->initialized )
	{
	hc->maxencodedurl = hc->maxdecodedurl = hc->maxprotocol =
	    hc->maxorigfilename = hc->maxexpnfilename = hc->maxencodings =
	    hc->maxpathinfo = hc->maxquery = hc->maxaltdir = hc->maxreferer =
	    hc->maxuseragent = hc->maxaccept = hc->maxaccepte =
	    hc->maxcookie = hc->maxcontenttype = 0;
	realloc_str( &hc->origfilename, &hc->maxorigfilename, 1 );
	realloc_str( &hc->expnfilename, &hc->maxexpnfilename, 0 );
	realloc_str( &hc->encodings, &hc->maxencodings, 0 );
	realloc_str( &hc->pathinfo, &hc->maxpathinfo, 0 );
	realloc_str( &hc->query, &hc->maxquery, 0 );
	realloc_str( &hc->altdir, &hc->maxaltdir, 0 );
	realloc_str( &hc->referer, &hc->maxreferer, 0 );
	realloc_str( &hc->useragent, &hc->maxuseragent, 0 );
	realloc_str( &hc->accept, &hc->maxaccept, 0 );
	realloc_str( &hc->accepte, &hc->maxaccepte, 0 );
	realloc_str( &hc->cookie, &hc->maxcookie, 0 );
	realloc_str( &hc->contenttype, &hc->maxcontenttype, 0 );
	hc->initialized = 1;
	}

    hc->hs = hs;

    /* Accept the new connection. */
    sz = sizeof(sin);
    hc->fd = accept( hs->fd, (struct sockaddr*) &sin, &sz );
    if ( hc->fd < 0 )
	{
	syslog( LOG_ERR, "accept - %m" );
	return -1;
	}
    hc->client_addr = sin.sin_addr;

    /* Open a read stream so that we can use fgets(), because it's much
    ** more efficient.  This has a couple of ramifications.  One, when
    ** we fclose() rfp, fd gets automatically close()d, so we don't want
    ** to close() it ourselves.  Two, when we get a POST request we will
    ** read part of the request data into the stdio buffer, so we need
    **an extra sub-process to send that to the CGI process.
    */
    hc->rfp = fdopen( hc->fd, "r" );
    if ( hc->rfp == (FILE*) 0 )
	{
	syslog( LOG_ERR, "fdopen - %m" );
	(void) close( hc->fd );
	return -1;
	}

    /* Set up an alarm to kill the read after a short timeout. */
    (void) alarm( READ_TIMELIMIT );

    /* Read request. */
    if ( fgets( buf, sizeof(buf), hc->rfp ) == (char*) 0 )
	{
	if ( ferror( hc->rfp ) && errno != EINTR && errno != ECONNRESET )
	    {
	    syslog( LOG_ERR, "fgets - %m" );
	    httpd_send_err( hc, 500, err500title, err500form, "" );
	    }
	else
	    httpd_send_err( hc, 408, err408title, err408form, "" );
	(void) fclose( hc->rfp );
	(void) alarm( 0 );
	return -1;
	}
    (void) alarm( 0 );

    /* Parse request. */
    hc->mime_flag = 0;
    method_str = buf;
    url = strpbrk( method_str, " \t\n\r" );
    if ( url == (char*) 0 )
	{
	httpd_send_err( hc, 400, err400title, err400form, buf );
	(void) fclose( hc->rfp );
	return -1;
	}
    *url++ = '\0';
    url += strspn( url, " \t\n\r" );
    protocol = strpbrk( url, " \t\n\r" );
    if ( protocol == (char*) 0 )
	protocol = "HTTP/0.9";
    else
	{
	*protocol++ = '\0';
	protocol += strspn( protocol, " \t\n\r" );
	if ( *protocol != '\0' )
	    {
	    hc->mime_flag = 1;
	    eol = strpbrk( protocol, " \t\n\r" );
	    if ( eol != (char*) 0 )
		*eol = '\0';
	    }
	}

    if ( strcmp( method_str, httpd_method_str( METHOD_GET ) ) == 0 )
	hc->method = METHOD_GET;
    else if ( strcmp( method_str, httpd_method_str( METHOD_HEAD ) ) == 0 )
	hc->method = METHOD_HEAD;
    else if ( strcmp( method_str, httpd_method_str( METHOD_POST ) ) == 0 )
	hc->method = METHOD_POST;
    else
	{
	httpd_send_err( hc, 501, err501title, err501form, method_str );
	(void) fclose( hc->rfp );
	return -1;
	}

    realloc_str( &hc->encodedurl, &hc->maxencodedurl, strlen( url ) );
    (void) strcpy( hc->encodedurl, url );
    realloc_str( &hc->decodedurl, &hc->maxdecodedurl, strlen( url ) );
    strdecode( hc->decodedurl, url );

    realloc_str( &hc->protocol, &hc->maxprotocol, strlen( protocol ) );
    (void) strcpy( hc->protocol, protocol );

    if ( hc->decodedurl[0] != '/' )
	{
	httpd_send_err( hc, 400, err400title, err400form, hc->encodedurl );
	(void) fclose( hc->rfp );
	return -1;
	}
    realloc_str(
	&hc->origfilename, &hc->maxorigfilename, strlen( hc->decodedurl ) );
    (void) strcpy( hc->origfilename, &hc->decodedurl[1] );
    /* Special case for top-level URL. */
    if ( hc->origfilename[0] == '\0' )
	(void) strcpy( hc->origfilename, "." );

    /* Extract query string from encoded URL. */
    hc->query[0] = '\0';
    cp = strchr( hc->encodedurl, '?' );
    if ( cp != (char*) 0 )
	{
	++cp;
	realloc_str( &hc->query, &hc->maxquery, strlen( cp ) );
	(void) strcpy( hc->query, cp );
	}
    /* And remove query from filename. */
    cp = strchr( hc->origfilename, '?' );
    if ( cp != (char*) 0 )
	*cp = '\0';

    /* Copy original filename to expanded filename. */
    realloc_str(
	&hc->expnfilename, &hc->maxexpnfilename, strlen( hc->origfilename ) );
    (void) strcpy( hc->expnfilename, hc->origfilename );

    /* Tilde mapping. */
    hc->altdir[0] = '\0';
    if ( hc->expnfilename[0] == '~' )
	{
	if ( ! tilde_map( hc ) )
	    {
	    httpd_send_err( hc, 500, err500title, err500form, hc->encodedurl );
	    (void) fclose( hc->rfp );
	    return -1;
	    }
	}

    /* Expand all symbolic links in the filename.  This also gives us
    ** any trailing non-existing components, for pathinfo.
    */
    cp = expand_symlinks( hc->expnfilename, &pi );
    if ( cp == (char*) 0 )
	{
	httpd_send_err( hc, 500, err500title, err500form, hc->encodedurl );
	(void) fclose( hc->rfp );
	return -1;
	}
    realloc_str( &hc->expnfilename, &hc->maxexpnfilename, strlen( cp ) );
    (void) strcpy( hc->expnfilename, cp );
    realloc_str( &hc->pathinfo, &hc->maxpathinfo, strlen( pi ) );
    (void) strcpy( hc->pathinfo, pi );

    /* Remove pathinfo stuff from the original filename too. */
    if ( hc->pathinfo[0] != '\0' )
	{
	int i;
	i = strlen( hc->origfilename ) - strlen( hc->pathinfo );
	if ( i > 0 && strcmp( &hc->origfilename[i], hc->pathinfo ) == 0 )
	    hc->origfilename[i - 1] = '\0';
	}

    /* If the expanded filename is an absolute path, check that it's still
    ** within the current directory or the alternate directory.
    */
    if ( hc->expnfilename[0] == '/' )
	{
	if ( strncmp(
		 hc->expnfilename, hc->hs->cwd, strlen( hc->hs->cwd ) ) == 0 )
	    /* Elide the current directory. */
	    (void) strcpy(
		hc->expnfilename, &hc->expnfilename[strlen( hc->hs->cwd )] );
	else if ( hc->altdir[0] != '\0' &&
		  ( strncmp(
		       hc->expnfilename, hc->altdir,
		       strlen( hc->altdir ) ) != 0 ||
		    ( hc->expnfilename[strlen( hc->altdir )] != '\0' &&
		      hc->expnfilename[strlen( hc->altdir )] != '/' ) ) )
	    {
	    httpd_send_err( hc, 403, err403title, err403form, hc->encodedurl );
	    (void) fclose( hc->rfp );
	    return -1;
	    }
	}

    hc->referer[0] = '\0';
    hc->useragent[0] = '\0';
    hc->accept[0] = '\0';
    hc->accepte[0] = '\0';
    hc->cookie[0] = '\0';
    hc->contenttype[0] = '\0';

    hc->file_fd = -1;
    hc->if_modified_since = (time_t) -1;
    hc->contentlength = -1;

    if ( hc->mime_flag )
	{
	/* Set up an the alarm again. */
	(void) alarm( READ_TIMELIMIT );

	/* Read any other headers in the request, particularly referer. */
	while ( fgets( buf, sizeof(buf), hc->rfp ) != (char*) 0 )
	    {
	    sz = strlen( buf );
	    while ( sz > 0 && ( buf[sz-1] == '\n' || buf[sz-1] == '\r' ) )
		buf[--sz] = '\0';
	    if ( sz == 0 )
		break;
	    if ( strncasecmp( buf, "Referer:", 8 ) == 0 )
		{
		cp = &buf[8];
		cp += strspn( cp, " \t" );
		realloc_str( &hc->referer, &hc->maxreferer, strlen( cp ) );
		(void) strcpy( hc->referer, cp );
		}
	    else if ( strncasecmp( buf, "User-Agent:", 11 ) == 0 )
		{
		cp = &buf[11];
		cp += strspn( cp, " \t" );
		realloc_str( &hc->useragent, &hc->maxuseragent, strlen( cp ) );
		(void) strcpy( hc->useragent, cp );
		}
	    else if ( strncasecmp( buf, "Accept:", 7 ) == 0 )
		{
		cp = &buf[7];
		cp += strspn( cp, " \t" );
		if ( hc->accept[0] != '\0' )
		    {
		    if ( strlen( hc->accept ) > 5000 )
			{
			syslog(
			    LOG_ERR, "%.80s way too much Accept: data",
			    inet_ntoa( hc->client_addr ) );
			continue;
			}
		    realloc_str(
			&hc->accept, &hc->maxaccept,
			strlen( hc->accept ) + 2 + strlen( cp ) );
		    (void) strcat( hc->accept, ", " );
		    }
		else
		    realloc_str( &hc->accept, &hc->maxaccept, strlen( cp ) );
		(void) strcat( hc->accept, cp );
		}
	    else if ( strncasecmp( buf, "Accept-Encoding:", 16 ) == 0 )
		{
		cp = &buf[16];
		cp += strspn( cp, " \t" );
		if ( hc->accepte[0] != '\0' )
		    {
		    if ( strlen( hc->accepte ) > 5000 )
			{
			syslog(
			    LOG_ERR, "%.80s way too much Accept-Encoding: data",
			    inet_ntoa( hc->client_addr ) );
			continue;
			}
		    realloc_str(
			&hc->accepte, &hc->maxaccepte,
			strlen( hc->accepte ) + 2 + strlen( cp ) );
		    (void) strcat( hc->accepte, ", " );
		    }
		else
		    realloc_str( &hc->accepte, &hc->maxaccepte, strlen( cp ) );
		(void) strcpy( hc->accepte, cp );
		}
	    else if ( strncasecmp( buf, "If-Modified-Since:", 18 ) == 0 )
		{
		cp = &buf[18];
		hc->if_modified_since = tdate_parse( cp );
		if ( hc->if_modified_since == (time_t) -1 )
		    syslog( LOG_DEBUG, "unparsable time: %.80s", cp );
		}
	    else if ( strncasecmp( buf, "Cookie:", 7 ) == 0 )
		{
		cp = &buf[7];
		cp += strspn( cp, " \t" );
		realloc_str( &hc->cookie, &hc->maxcookie, strlen( cp ) );
		(void) strcpy( hc->cookie, cp );
		}
	    else if ( strncasecmp( buf, "Content-Type:", 13 ) == 0 )
		{
		cp = &buf[13];
		cp += strspn( cp, " \t" );
		realloc_str( &hc->contenttype, &hc->maxcontenttype, strlen( cp ) );
		(void) strcpy( hc->contenttype, cp );
		}
	    else if ( strncasecmp( buf, "Content-Length:", 15 ) == 0 )
		{
		cp = &buf[15];
		hc->contentlength = atoi( cp );
		}
	    else if ( strncasecmp( buf, "Accept-Charset:", 15 ) == 0 ||
	              strncasecmp( buf, "Accept-Language:", 16 ) == 0 ||
	              strncasecmp( buf, "Authorization:", 14 ) == 0 ||
	              strncasecmp( buf, "Cache-Control:", 14 ) == 0 ||
	              strncasecmp( buf, "Cache-Info:", 11 ) == 0 ||
	              strncasecmp( buf, "Charge-To:", 10 ) == 0 ||
	              strncasecmp( buf, "Client-ip:", 10 ) == 0 ||
	              strncasecmp( buf, "Connection:", 11 ) == 0 ||
	              strncasecmp( buf, "Date:", 5 ) == 0 ||
	              strncasecmp( buf, "Extension:", 10 ) == 0 ||
	              strncasecmp( buf, "Forwarded:", 10 ) == 0 ||
	              strncasecmp( buf, "From:", 5 ) == 0 ||
	              strncasecmp( buf, "Host:", 5 ) == 0 ||
	              strncasecmp( buf, "HTTP-Version:", 13 ) == 0 ||
	              strncasecmp( buf, "Message-ID:", 11 ) == 0 ||
	              strncasecmp( buf, "MIME-Version:", 13 ) == 0 ||
	              strncasecmp( buf, "Pragma:", 7 ) == 0 ||
	              strncasecmp( buf, "Proxy-agent:", 12 ) == 0 ||
	              strncasecmp( buf, "Proxy-Connection:", 17 ) == 0 ||
	              strncasecmp( buf, "Security-Scheme:", 16 ) == 0 ||
	              strncasecmp( buf, "Session-ID:", 11 ) == 0 ||
	              strncasecmp( buf, "UA-color:", 9 ) == 0 ||
	              strncasecmp( buf, "UA-CPU:", 7 ) == 0 ||
	              strncasecmp( buf, "UA-Disp:", 8 ) == 0 ||
	              strncasecmp( buf, "UA-OS:", 6 ) == 0 ||
	              strncasecmp( buf, "UA-pixels:", 10 ) == 0 ||
	              strncasecmp( buf, "User:", 5 ) == 0 ||
	              strncasecmp( buf, "Via:", 4 ) == 0 ||
	              strncasecmp( buf, "X-", 2 ) == 0 )
		; /* ignore */
	    else
		syslog( LOG_DEBUG, "unknown MIME header: %.80s", buf );
	    }
	(void) alarm( 0 );
	}

    return 0;
    }


void
httpd_close_conn( httpd_conn* hc )
    {
    if ( hc->file_fd >= 0 )
	{
	(void) close( hc->file_fd );
	hc->file_fd = -1;
	}
    if ( hc->rfp != (FILE*) 0 )
	{
	(void) fclose( hc->rfp );
	hc->rfp = (FILE*) 0;
	hc->fd = -1;
	}
    else if ( hc->fd >= 0 )
	{
	(void) close( hc->fd );
	hc->fd = -1;
	}
    }


/* Figures out MIME encodings and type based on the filename.  Multiple
** encodings are separated by semicolons.
*/
static void
figure_mime( httpd_conn* hc )
    {
    int i, j, k, l;
    int got_enc;
    struct table {
	char* ext;
	char* val;
	};
    static struct table enc_tab[] = {
#include "mime_encodings.h"
	};
    static struct table typ_tab[] = {
#include "mime_types.h"
	};

    /* Look at the extensions on hc->expnfilename from the back forwards. */
    hc->encodings[0] = '\0';
    i = strlen( hc->expnfilename );
    for (;;)
	{
	j = i;
	for (;;)
	    {
	    --i;
	    if ( i <= 0 )
		{
		/* No extensions left. */
		hc->type = "text/plain";
		return;
		}
	    if ( hc->expnfilename[i] == '.' )
		break;
	    }
	/* Found an extension. */
	got_enc = 0;
	for ( k = 0; k < sizeof(enc_tab)/sizeof(*enc_tab); ++k )
	    {
	    l = strlen( enc_tab[k].ext );
	    if ( l == j - i - 1 &&
		 strncasecmp( &hc->expnfilename[i+1], enc_tab[k].ext, l ) == 0 )
		{
		realloc_str(
		    &hc->encodings, &hc->maxencodings,
		    strlen( enc_tab[k].val ) + 1 );
		if ( hc->encodings[0] != '\0' )
		    (void) strcat( hc->encodings, ";" );
		(void) strcat( hc->encodings, enc_tab[k].val );
		got_enc = 1;
		}
	    }
	if ( ! got_enc )
	    {
	    /* No encoding extension found - time to try type extensions. */
	    for ( k = 0; k < sizeof(typ_tab)/sizeof(*typ_tab); ++k )
		{
		l = strlen( typ_tab[k].ext );
		if ( l == j - i - 1 &&
		     strncasecmp(
			 &hc->expnfilename[i+1], typ_tab[k].ext, l ) == 0 )
		    {
		    hc->type = typ_tab[k].val;
		    return;
		    }
		}
	    /* No recognized type extension found - return default. */
	    hc->type = "text/plain";
	    return;
	    }
	}
    }


#ifdef CGI_TIMELIMIT
static void
cgi_kill2( ClientData client_data )
    {
    pid_t pid;

    pid = (pid_t) client_data.i;
    if ( kill( pid, SIGKILL ) == 0 )
	syslog( LOG_ERR, "hard-killed CGI process %d", pid );
    }

static void
cgi_kill( ClientData client_data )
    {
    pid_t pid;
    struct timeval tv;

    pid = (pid_t) client_data.i;
    if ( kill( pid, SIGINT ) == 0 )
	{
	syslog( LOG_ERR, "killed CGI process %d", pid );
	/* In case this isn't enough, schedule an uncatchable kill. */
	(void) gettimeofday( &tv, (struct timezone*) 0 );
	(void) tmr_create( &tv, cgi_kill2, client_data, 5 * 1000000L, 0 );
	}
    }
#endif /* CGI_TIMELIMIT */


/* qsort comparison routine - declared old-style on purpose, for portability. */
static int
name_compare( a, b )
    char** a;
    char** b;
    {
    return strcmp( *a, *b );
    }


static off_t
ls( httpd_conn* hc )
    {
    DIR* dirp;
    struct dirent* de;
    static int maxnames = 0;
    int nnames;
    static char* names;
    static char** nameptrs;
    static char* name;
    static int maxname = 0;
    static char* rname;
    static int maxrname = 0;
    FILE* fp;
    int i, r;
    struct stat sb;
    struct stat lsb;
    char modestr[20];
    char* linkprefix;
    char link[MAXPATHLEN];
    int linklen;
    char* fileclass;
    time_t now;
    char* timestr;
    struct timeval tv;
    ClientData client_data;

    dirp = opendir( hc->expnfilename );
    if ( dirp == (DIR*) 0 )
	{
	syslog( LOG_ERR, "opendir - %m" );
	httpd_send_err( hc, 404, err404title, err404form, hc->encodedurl );
	return -1;
	}

    send_mime(
	hc, 200, ok200title, "", "", "text/html", -1, hc->sb.st_mtime );
    hc->bytes = 0;
    if ( hc->method == METHOD_HEAD )
	closedir( dirp );
    else if ( hc->method == METHOD_GET )
	{
	r = fork( );
	if ( r < 0 )
	    {
	    syslog( LOG_ERR, "fork - %m" );
	    httpd_send_err( hc, 500, err500title, err500form, hc->encodedurl );
	    return -1;
	    }
	if ( r == 0 )
	    {
	    /* Child process. */

#ifdef CGI_NICE
	    /* Set priority. */
	    (void) nice( CGI_NICE );
#endif /* CGI_NICE */

	    /* Open a stdio stream so that we can use fprintf, which is more
	    ** efficient than a bunch of separate write()s.  We don't have
	    ** to worry about double closes or file descriptor leaks cause
	    ** we're in a subprocess.
	    */
	    fp = fdopen( hc->fd, "w" );
	    if ( fp == (FILE*) 0 )
		{
		syslog( LOG_ERR, "fdopen - %m" );
		httpd_send_err(
		    hc, 500, err500title, err500form, hc->encodedurl );
		closedir( dirp );
		return -1;
		}

	    (void) fprintf( fp, "\
<HTML><HEAD><TITLE>Index of %.80s</TITLE></HEAD>\n\
<BODY>\n\
<H2>Index of %.80s</H2>\n\
<PRE>\n\
mode  links  bytes  last-changed  name\n\
<HR>",
		hc->encodedurl, hc->encodedurl );

	    /* Read in names. */
	    nnames = 0;
	    while ( ( de = readdir( dirp ) ) != 0 )	/* dirent or direct */
		{
		if ( nnames >= maxnames )
		    {
		    if ( maxnames == 0 )
			{
			maxnames = 100;
			names = NEW( char, maxnames * MAXPATHLEN );
			nameptrs = NEW( char*, maxnames );
			}
		    else
			{
			maxnames *= 2;
			names = RENEW( names, char, maxnames * MAXPATHLEN );
			nameptrs = RENEW( nameptrs, char*, maxnames );
			}
		    if ( names == (char*) 0  || nameptrs == (char**) 0 )
			{
			syslog( LOG_ERR, "out of memory" );
			exit( 1 );
			}
		    for ( i = 0; i < maxnames; ++i )
			nameptrs[i] = &names[i * MAXPATHLEN];
		    }
		(void) strcpy( nameptrs[nnames], de->d_name );
		++nnames;
		}
	    closedir( dirp );

	    /* Sort the names. */
	    qsort( nameptrs, nnames, sizeof(*nameptrs), name_compare );

	    /* Generate output. */
	    for ( i = 0; i < nnames; ++i )
		{
		realloc_str(
		    &name, &maxname,
		    strlen( hc->expnfilename ) + 1 + strlen( nameptrs[i] ) );
		realloc_str(
		    &rname, &maxrname,
		    strlen( hc->origfilename ) + 1 + strlen( nameptrs[i] ) );
		if ( hc->expnfilename[0] == '\0' ||
		     strcmp( hc->expnfilename, "." ) == 0 )
		    {
		    (void) strcpy( name, nameptrs[i] );
		    (void) strcpy( rname, nameptrs[i] );
		    }
		else
		    {
		    (void) sprintf(
			name, "%s/%s", hc->expnfilename, nameptrs[i] );
		    (void) sprintf(
			rname, "%s%s", hc->origfilename, nameptrs[i] );
		    }
		if ( stat( name, &sb ) < 0 || lstat( name, &lsb ) < 0 )
		    continue;
		linkprefix = "";
		link[0] = '\0';
		/* Break down mode word.  First the file type. */
		switch ( lsb.st_mode & S_IFMT )
		    {
		    case S_IFIFO:  modestr[0] = 'p'; break;
		    case S_IFCHR:  modestr[0] = 'c'; break;
		    case S_IFDIR:  modestr[0] = 'd'; break;
		    case S_IFBLK:  modestr[0] = 'b'; break;
		    case S_IFREG:  modestr[0] = '-'; break;
		    case S_IFSOCK: modestr[0] = 's'; break;
		    case S_IFLNK:  modestr[0] = 'l';
		    linklen = readlink( name, link, sizeof(link) );
		    if ( linklen != -1 )
			{
			link[linklen] = '\0';
			linkprefix = " -> ";
			}
		    break;
		    default:       modestr[0] = '?'; break;
		    }
		/* Now the world permissions.  Owner and group permissions
		** are not of interest to web clients.
		*/
		modestr[1] = ( lsb.st_mode & S_IROTH ) ? 'r' : '-';
		modestr[2] = ( lsb.st_mode & S_IWOTH ) ? 'w' : '-';
		modestr[3] = ( lsb.st_mode & S_IXOTH ) ? 'x' : '-';
		modestr[4] = '\0';

		/* We also leave out the owner and group name, they are
		** also not of interest to web clients.  Plus if we're
		** running under chroot(), they would require a copy
		** of /etc/passwd and /etc/group, which we want to avoid.
		*/

		/* Get time string. */
		now = time( (time_t*) 0 );
		timestr = ctime( &lsb.st_mtime );
		timestr[ 0] = timestr[ 4];
		timestr[ 1] = timestr[ 5];
		timestr[ 2] = timestr[ 6];
		timestr[ 3] = ' ';
		timestr[ 4] = timestr[ 8];
		timestr[ 5] = timestr[ 9];
		timestr[ 6] = ' ';
		if ( now - lsb.st_mtime > 60*60*24*182 )	/* 1/2 year */
		    {
		    timestr[ 7] = ' ';
		    timestr[ 8] = timestr[20];
		    timestr[ 9] = timestr[21];
		    timestr[10] = timestr[22];
		    timestr[11] = timestr[23];
		    }
		else
		    {
		    timestr[ 7] = timestr[11];
		    timestr[ 8] = timestr[12];
		    timestr[ 9] = ':';
		    timestr[10] = timestr[14];
		    timestr[11] = timestr[15];
		    }
		timestr[12] = '\0';

		/* The ls -F file class. */
		switch ( sb.st_mode & S_IFMT )
		    {
		    case S_IFDIR:  fileclass = "/"; break;
		    case S_IFSOCK: fileclass = "="; break;
		    case S_IFLNK:  fileclass = "@"; break;
		    default:
		    fileclass = ( sb.st_mode & S_IXOTH ) ? "*" : "";
		    break;
		    }

		/* And print. */
		(void)  fprintf( fp,
		   "%s %3ld  %8ld  %s  <A HREF=\"/%.500s%s\">%.80s</A>%s%s%s\n",
		    modestr, (long) lsb.st_nlink, (long) lsb.st_size, timestr,
		    rname, S_ISDIR(sb.st_mode) ? "/" : "", nameptrs[i],
		    linkprefix, link, fileclass );
		}

	    (void) fprintf( fp, "</PRE></BODY></HTML>\n" );
	    (void) fclose( fp );
	    exit( 0 );
	    }

	/* Parent process. */
	closedir( dirp );
	syslog( LOG_DEBUG, "LS pid is %d", r );
#ifdef CGI_TIMELIMIT
	/* Schedule a kill for the child process, in case it runs too long */
	(void) gettimeofday( &tv, (struct timezone*) 0 );
	client_data.i = r;
	(void) tmr_create(
	    &tv, cgi_kill, client_data, CGI_TIMELIMIT * 1000000L, 0 );
#endif /* CGI_TIMELIMIT */
	hc->status = 200;
	hc->bytes = CGI_BYTECOUNT;
	}
    else
	{
	httpd_send_err(
	    hc, 501, err501title, err501form, httpd_method_str( hc->method ) );
	return -1;
	}

    return 0;
    }


static char*
build_env( char* fmt, char* arg )
    {
    char* cp;
    int size;
    static char* buf;
    static int maxbuf = 0;

    size = strlen( fmt ) + strlen( arg );
    if ( size > maxbuf )
        realloc_str( &buf, &maxbuf, size );
    (void) sprintf( buf, fmt, arg );
    cp = strdup( buf );
    if ( cp == (char*) 0 )
        {
        syslog( LOG_ERR, "out of memory" );
        exit( 1 );
        }
    return cp;
    }


#ifdef SERVER_NAME_LIST
static char*
hostname_map( char* hostname )
    {
    int len, n;
    static char* list[] = { SERVER_NAME_LIST };

    len = strlen( hostname );
    for ( n = sizeof(list) / sizeof(*list) - 1; n >= 0; --n )
	if ( strncasecmp( hostname, list[n], len ) == 0 )
	    if ( list[n][len] == '/' )	/* check in case of a substring match */
		return &list[n][len + 1];
    return (char*) 0;
    }
#endif /* SERVER_NAME_LIST */


/* Set up environment variables. Be real careful here to avoid
** letting malicious clients overrun a buffer.  We don't have
** to worry about freeing stuff since we're a sub-process.
*/
static char**
make_envp( httpd_conn* hc )
    {
    static char* envp[50];
    int envn;
    char* cp;
    char buf[256];

    envn = 0;
    envp[envn++] = build_env( "PATH=%s", CGI_PATH );
    envp[envn++] = build_env( "SERVER_SOFTWARE=%s", SERVER_SOFTWARE );
    cp = hc->hs->hostname;
#ifdef SERVER_NAME_LIST
    if ( cp == (char*) 0 && gethostname( buf, sizeof(buf) ) >= 0 )
	cp = hostname_map( buf );
#endif /* SERVER_NAME_LIST */
    if ( cp == (char*) 0 )
	{
#ifdef SERVER_NAME
	cp = SERVER_NAME;
#else /* SERVER_NAME */
	if ( gethostname( buf, sizeof(buf) ) >= 0 )
	    cp = buf;
#endif /* SERVER_NAME */
	}
    if ( cp != (char*) 0 )
	envp[envn++] = build_env( "SERVER_NAME=%s", cp );
    envp[envn++] = "GATEWAY_INTERFACE=CGI/1.1";
    envp[envn++] = build_env("SERVER_PROTOCOL=%s", hc->protocol);
    (void) sprintf( buf, "%d", hc->hs->port );
    envp[envn++] = build_env( "SERVER_PORT=%s", buf );
    envp[envn++] = build_env(
	"REQUEST_METHOD=%s", httpd_method_str( hc->method ) );
    if ( hc->pathinfo[0] != '\0' )
	{
	char* cp2;
	envp[envn++] = build_env( "PATH_INFO=/%s", hc->pathinfo );
	cp2 = NEW( char, strlen( hc->hs->cwd ) + strlen( hc->pathinfo ) );
	if ( cp2 != (char*) 0 )
	    {
	    (void) sprintf( cp2, "%s%s", hc->hs->cwd, hc->pathinfo );
	    envp[envn++] = build_env( "PATH_TRANSLATED=%s", cp2 );
	    }
	}
    envp[envn++] = build_env( "SCRIPT_NAME=/%s", hc->origfilename );
    if ( hc->query[0] != '\0')
	envp[envn++] = build_env( "QUERY_STRING=%s", hc->query );
    envp[envn++] = build_env( "REMOTE_ADDR=%s", inet_ntoa( hc->client_addr ) );
    if ( hc->referer[0] != '\0' )
	envp[envn++] = build_env( "HTTP_REFERER=%s", hc->referer );
    if ( hc->useragent[0] != '\0' )
	envp[envn++] = build_env( "HTTP_USER_AGENT=%s", hc->useragent );
    if ( hc->accept[0] != '\0' )
	envp[envn++] = build_env( "HTTP_ACCEPT=%s", hc->accept );
    if ( hc->accepte[0] != '\0' )
	envp[envn++] = build_env( "HTTP_ACCEPT_ENCODING=%s", hc->accepte );
    if ( hc->cookie[0] != '\0' )
	envp[envn++] = build_env( "HTTP_COOKIE=%s", hc->cookie );
    if ( hc->contenttype[0] != '\0' )
	envp[envn++] = build_env( "CONTENT_TYPE=%s", hc->contenttype );
    if ( hc->contentlength != -1 )
	{
	(void) sprintf( buf, "%d", hc->contentlength );
	envp[envn++] = build_env( "CONTENT_LENGTH=%s", buf );
	}
    if ( getenv( "TZ" ) != (char*) 0 )
	envp[envn++] = build_env( "TZ=%s", getenv( "TZ" ) );

    envp[envn] = (char*) 0;
    return envp;
    }


/* Set up argument vector.  Again, we don't have to worry about freeing stuff
** since we're a sub-process.  This gets done after make_envp() because we
** scribble on hc->query.
*/
static char**
make_argp( httpd_conn* hc )
    {
    char** argp;
    int argn;
    char* cp1;
    char* cp2;

    /* By allocating an arg slot for every character in the query, plus
    ** one for the filename and one for the NULL, we are guaranteed to
    ** have enough.  We could actually use strlen/2.
    */
    argp = NEW( char*, strlen( hc->query ) + 2 );
    if ( argp == (char**) 0 )
	return (char**) 0;

    argp[0] = strrchr( hc->expnfilename, '/' );
    if ( argp[0] != (char*) 0 )
	++argp[0];
    else
	argp[0] = hc->expnfilename;

    argn = 1;
    /* According to the CGI spec at http://hoohoo.ncsa.uiuc.edu/cgi/cl.html,
    ** "The server should search the query information for a non-encoded =
    ** character to determine if the command line is to be used, if it finds
    ** one, the command line is not to be used."
    */
    if ( strchr( hc->query, '=' ) == (char*) 0 )
	{
	for ( cp1 = cp2 = hc->query; *cp2 != '\0'; ++cp2 )
	    {
	    if ( *cp2 == '+' )
		{
		*cp2 = '\0';
		strdecode( cp1, cp1 );
		argp[argn++] = cp1;
		cp1 = cp2 + 1;
		}
	    }
	if ( cp2 != cp1 )
	    {
	    strdecode( cp1, cp1 );
	    argp[argn++] = cp1;
	    }
	}

    argp[argn] = (char*) 0;
    return argp;
    }


/* This routine is used only for POST requests.  It reads the data
** from the request and sends it to the child process.  The only reason
** we need to do it this way instead of just letting the child read
** directly is that we have already read part of the data into our
** stdio buffer.
*/
static void
cgi_interpose( httpd_conn* hc, int wfd )
    {
    char buf[1024];
    int i, c;

    for ( i = c = 0; c <= hc->contentlength && !feof( hc->rfp ); ++i, ++c )
	{
	if ( i >= sizeof(buf) )
	    {
	    (void) write( wfd, buf, i );
	    i = 0;
	    }
	buf[i] = fgetc( hc->rfp );
	}
    if ( i > 0 )
	(void) write( wfd, buf, i );
    exit( 0 );
    }


/* CGI child process. */
static void
cgi_child( httpd_conn* hc )
    {
    int r, fd;
    char** argp;
    char** envp;
    char http_head[] = "HTTP/1.0 200 OK\n";
    char* binary;
    char* directory;

    /* Set up stdin. */
    if ( hc->method == METHOD_POST )
	{
	/* For POSTs we have to set up a pipe from an interposer process
	** for stdin.
	*/
	int p[2];

	if ( pipe( p ) < 0 )
	    {
	    syslog( LOG_ERR, "pipe - %m" );
	    httpd_send_err(
		hc, 500, err500title, err500form, hc->encodedurl );
	    exit( 1 );
	    }
	r = fork( );
	if ( r < 0 )
	    {
	    syslog( LOG_ERR, "fork - %m" );
	    httpd_send_err(
		hc, 500, err500title, err500form, hc->encodedurl );
	    exit( 1 );
	    }
	if ( r == 0 )
	    {
	    /* Interposer process. */
	    (void) close( p[0] );
	    cgi_interpose( hc, p[1] );
	    }
	(void) close( p[1] );
	(void) dup2( p[0], STDIN_FILENO );
	}
    else
	{
	/* Otherwise, the request socket is stdin and we probably won't
	** read from it.  In theory we could set it to /dev/null.
	*/
	(void) dup2( hc->fd, STDIN_FILENO );
	}

    /* The response socket always becomes stdout and stderr. */
    (void) dup2( hc->fd, STDOUT_FILENO );
    (void) dup2( hc->fd, STDERR_FILENO );

    /* Close all other descriptors so that the CGI program can't
    ** mess with them.  Close the syslog descriptor first or else
    ** syslog() will be writing messages to a closed file.
    */
    closelog();
    for ( fd = httpd_get_nfiles() - 1; fd > 2; --fd )
	(void) close( fd );

    /* Make the environment vector. */
    envp = make_envp( hc );

    /* Make the argument vector. */
    argp = make_argp( hc );

    /* Be minimally compatible with the parsed-headers vs. nph- crap;
    ** but never send header for HTTP/0.9
    */
    if ( strncmp( argp[0], "nph-", 4 ) != 0 && hc->mime_flag )
	(void) write( STDOUT_FILENO, http_head, sizeof(http_head) - 1 );

#ifdef CGI_NICE
    /* Set priority. */
    (void) nice( CGI_NICE );
#endif /* CGI_NICE */

    /* Split the program into directory and binary, so we can chdir()
    ** to the program's own directory.  This isn't in the CGI 1.1
    ** spec, but it's what most other HTTP servers do.
    */
    directory = strdup( hc->expnfilename );
    if ( directory == (char*) 0 )
	binary = hc->expnfilename;	/* ignore errors */
    else
	{
	binary = strrchr( directory, '/' );
	if ( binary == (char*) 0 )
	    binary = hc->expnfilename;
	else
	    {
	    *binary++ = '\0';
	    (void) chdir( directory );	/* ignore errors */
	    }
	}

    /* Run the program. */
    (void) execve( binary, argp, envp );

    /* Something went wrong. */
    syslog( LOG_ERR, "execve %s - %m", hc->expnfilename );
    httpd_send_err( hc, 500, err500title, err500form, hc->encodedurl );
    exit( 1 );
    }


static off_t
cgi( httpd_conn* hc )
    {
    int r;
    struct timeval tv;
    ClientData client_data;

    if ( hc->method == METHOD_HEAD )
	{
	/* HEAD doesn't make much sense for CGI.  We'll just make up a header
	** and send it back.
	*/
	send_mime( hc, 200, ok200title, "", "", "text/plain", -1, 0 );
	hc->bytes = 0;
	}
    else if ( hc->method == METHOD_GET || hc->method == METHOD_POST )
	{
	r = fork( );
	if ( r < 0 )
	    {
	    syslog( LOG_ERR, "fork - %m" );
	    httpd_send_err( hc, 500, err500title, err500form, hc->encodedurl );
	    return -1;
	    }
	if ( r == 0 )
	    cgi_child( hc );

	/* Parent process. */
	syslog( LOG_DEBUG, "CGI pid is %d", r );
#ifdef CGI_TIMELIMIT
	/* Schedule a kill for the child process, in case it runs too long */
	(void) gettimeofday( &tv, (struct timezone*) 0 );
	client_data.i = r;
	(void) tmr_create(
	    &tv, cgi_kill, client_data, CGI_TIMELIMIT * 1000000L, 0 );
#endif /* CGI_TIMELIMIT */
	hc->status = 200;
	hc->bytes = CGI_BYTECOUNT;
	}
    else
	{
	httpd_send_err(
	    hc, 501, err501title, err501form, httpd_method_str( hc->method ) );
	return -1;
	}

    return 0;
    }


static int
really_start_request( httpd_conn* hc )
    {
    static char* indexname;
    static int maxindexname = 0;
    int len;
    char* cp;
    char* pi;

    if ( hc->method != METHOD_GET && hc->method != METHOD_HEAD &&
	 hc->method != METHOD_POST )
	{
	httpd_send_err(
	    hc, 501, err501title, err501form, httpd_method_str( hc->method ) );
	return -1;
	}

    /* Stat the file. */
    if ( stat( hc->expnfilename, &hc->sb ) < 0 )
	{
	httpd_send_err( hc, 500, err500title, err500form, hc->encodedurl );
	return -1;
	}

    /* Is it world-readable?  We check explicitly instead of just
    ** trying to open it, so that no one ever gets surprised by
    ** a file that's not set world-readable and yet somehow is
    ** readable by the HTTP server and therefore the *whole* world.
    */
    if ( ! ( hc->sb.st_mode & S_IROTH ) )
	{
	httpd_send_err( hc, 403, err403title, err403form, hc->encodedurl );
	return -1;
	}

    /* Is it a directory? */
    if ( S_ISDIR(hc->sb.st_mode) )
	{
	/* Special handling for directory URLs that don't end in a slash.
	** We send back an explicit redirect with the slash, because
	** otherwise many clients can't build relative URLs properly.
	*/
	if ( hc->decodedurl[strlen( hc->decodedurl ) - 1] != '/' )
	    {
	    send_dirredirect( hc );
	    return -1;
	    }

	/* Check for an index.html file. */
	realloc_str(
	    &indexname, &maxindexname,
	    strlen( hc->expnfilename ) + 1 + sizeof(INDEX_NAME) );
	(void) strcpy( indexname, hc->expnfilename );
	len = strlen( indexname );
	if ( len == 0 || indexname[len - 1] != '/' )
	    (void) strcat( indexname, "/" );
	(void) strcat( indexname, INDEX_NAME );
	if ( stat( indexname, &hc->sb ) < 0 )
	    {
	    /* Nope, no index.html, so it's an actual directory request.
	    ** But if there's pathinfo, it's just a non-existent file.
	    */
	    if ( hc->pathinfo[0] != '\0' )
		{
		httpd_send_err(
		    hc, 404, err404title, err404form, hc->encodedurl );
		return -1;
		}
	    return ls( hc );
	    }
	/* Expand symlinks again.  More pathinfo means something went wrong. */
	cp = expand_symlinks( indexname, &pi );
	if ( cp == (char*) 0 || pi[0] != '\0' )
	    {
	    httpd_send_err( hc, 500, err500title, err500form, hc->encodedurl );
	    return -1;
	    }
	realloc_str( &hc->expnfilename, &hc->maxexpnfilename, strlen( cp ) );
	(void) strcpy( hc->expnfilename, cp );

	/* Now, is the index.html version world-readable? */
	if ( ! ( hc->sb.st_mode & S_IROTH ) )
	    {
	    httpd_send_err( hc, 403, err403title, err403form, hc->encodedurl );
	    return -1;
	    }
	}

    /* Is it world-executable and in the CGI area? */
    if ( hc->hs->cgi_pattern != (char*) 0 &&
	 ( hc->sb.st_mode & S_IXOTH ) &&
	 match( hc->hs->cgi_pattern, hc->expnfilename ) )
	return cgi( hc );

    /* It's not CGI.  If it's executable or there's pathinfo, someone's
    ** trying to either serve or run a non-CGI file as CGI.   Either case
    ** is prohibited.
    */
    if ( ( hc->sb.st_mode & S_IXOTH ) || hc->pathinfo[0] != '\0' )
	{
	httpd_send_err( hc, 403, err403title, err403form, hc->encodedurl );
	return -1;
	}

    /* It's just a regular old file.  Send it. */
    hc->file_fd = open( hc->expnfilename, O_RDONLY );
    if ( hc->file_fd < 0 )
	{
	if ( errno == ENOENT )
	    httpd_send_err( hc, 404, err404title, err404form, hc->encodedurl );
	else if ( errno == EACCES )
	    httpd_send_err( hc, 403, err403title, err403form, hc->encodedurl );
	else
	    {
	    syslog( LOG_ERR, "open - %m" );
	    httpd_send_err( hc, 500, err500title, err500form, hc->encodedurl );
	    }
	return -1;
	}

    figure_mime( hc );

    /* Handle If-Modified-Since. */
    if ( hc->if_modified_since != (time_t) -1 &&
	 hc->if_modified_since == hc->sb.st_mtime )
	{
	send_mime(
	    hc, 304, err304title, hc->encodings, "", hc->type, hc->sb.st_size,
	    hc->sb.st_mtime );
	hc->method = METHOD_HEAD;
	}
    else
	send_mime(
	    hc, 200, ok200title, hc->encodings, "", hc->type, hc->sb.st_size,
	    hc->sb.st_mtime );

    if ( hc->method == METHOD_HEAD )
	{
	hc->bytes = 0;
	(void) close( hc->file_fd );
	hc->file_fd = -1;
	}
    else
	hc->bytes = hc->sb.st_size;

    return 0;
    }


int
httpd_start_request( httpd_conn* hc )
    {
    int r;
    char bytes[40];
    char* ref;
    char* ua;

    r = really_start_request( hc );


    /* This is a slightly modified version of the CERN Combined Common
    ** Log Format - the difference is that we leave out the date, because
    ** syslogd puts it in.  A simple sed command can turn the results
    ** into true CERN format.
    */
    if ( (long) hc->bytes >= 0 )
	(void) sprintf( bytes, "%ld", (long) hc->bytes );
    else
	(void) strcpy( bytes, "-" );
#ifdef LOG_REFERER_ALWAYS
    ref = hc->referer;
#else /* LOG_REFERER_ALWAYS */
    ref = "";
#endif /* LOG_REFERER_ALWAYS */
#ifdef LOG_USERAGENT_ALWAYS
    ua = hc->useragent;
#else /* LOG_USERAGENT_ALWAYS */
    ua = "";
#endif /* LOG_USERAGENT_ALWAYS */
    syslog(
	LOG_INFO, "%.80s - - \"%.80s %.200s\" %d %s \"%.80s\" \"%.80s\"",
	inet_ntoa( hc->client_addr ), httpd_method_str( hc->method ),
	hc->encodedurl, hc->status, bytes, ref, ua );

    return r;
    }
